import datetime as dt
from contextlib import suppress
from enum import Enum as _Enum
from typing import Any
from typing import Generic
from typing import TypeVar

from beartype import beartype
from click import Context
from click import ParamType
from click import Parameter
from click import option

from utilities.logging import LogLevel


class Date(ParamType):
    """A date-valued parameter."""

    name = "date"

    @beartype
    def convert(
        self, value: Any, param: Parameter | None, ctx: Context | None
    ) -> dt.date:
        with suppress(ValueError):
            return dt.date.fromisoformat(value)
        with suppress(ValueError):
            return dt.datetime.strptime(value, "%Y%m%d").date()
        self.fail(f"Unable to parse {value}", param, ctx)


class DateTime(ParamType):
    """A datetime-valued parameter."""

    name = "datetime"

    @beartype
    def convert(
        self, value: Any, param: Parameter | None, ctx: Context | None
    ) -> dt.date:
        with suppress(ValueError):
            return dt.datetime.fromisoformat(value)
        with suppress(ValueError):
            return dt.datetime.strptime(value, "%Y%m%d%H%M%S")
        self.fail(f"Unable to parse {value}", param, ctx)


_E = TypeVar("_E", bound=_Enum)


class Enum(ParamType, Generic[_E]):
    """An enum-valued parameter."""

    name = "enum"

    @beartype
    def __init__(self, enum: type[_E]) -> None:
        super().__init__()
        self._enum = enum

    @beartype
    def convert(
        self, value: Any, param: Parameter | None, ctx: Context | None
    ) -> _E:
        els = {el for el in self._enum if el.name.lower() == value.lower()}
        with suppress(ValueError):
            (el,) = els
            return el
        self.fail(f"Unable to parse {value}", param, ctx)


log_level_option = option(
    "--log-level",
    type=Enum(LogLevel),
    default=LogLevel.INFO,
    show_default=True,
    help="The logging level",
)



import datetime as dt
from collections.abc import Iterable
from typing import Any
from typing import TypeVar

from beartype import beartype
from cattrs import Converter
from typed_settings import default_converter
from typed_settings import default_loaders
from typed_settings import load_settings as _load_settings

from utilities.pathlib import PathLike


_T = TypeVar("_T")


@beartype
def load_settings(
    cls: type[_T], appname: str, /, *, config_files: Iterable[PathLike] = ()
) -> _T:
    """Load a settings object with the extended converter."""

    loaders = default_loaders(appname, config_files=config_files)
    converter = _make_converter()
    return _load_settings(cls, loaders, converter=converter)


@beartype
def _make_converter() -> Converter:
    """Extend the default converter."""

    converter = default_converter()
    converter.register_structure_hook(dt.date, _to_date)
    converter.register_structure_hook(dt.time, _to_time)
    converter.register_structure_hook(dt.timedelta, _to_timedelta)
    return converter


@beartype
def _to_date(value: Any, _: type = dt.date) -> dt.date:
    """Convert a value to a date."""

    if isinstance(value, dt.date):
        return value
    elif isinstance(value, str):
        return dt.date.fromisoformat(value)
    else:
        raise TypeError(type(value))


@beartype
def _to_time(value: Any, _: type = dt.time) -> dt.time:
    """Convert a value to a time."""

    if isinstance(value, dt.time):
        return value
    elif isinstance(value, str):
        return dt.time.fromisoformat(value)
    else:
        raise TypeError(type(value))


@beartype
def _to_timedelta(value: Any, _: type = dt.timedelta) -> dt.timedelta:
    """Convert a value to a timedelta."""

    if isinstance(value, dt.timedelta):
        return value
    elif isinstance(value, str):
        as_date = dt.datetime.strptime(value, "%H:%M:%S")
        return dt.timedelta(
            hours=as_date.hour, minutes=as_date.minute, seconds=as_date.second
        )
    else:
        raise TypeError(type(value))



from pathlib import Path
from string import ascii_letters
from zipfile import ZipFile

from hypothesis import given
from hypothesis.strategies import sampled_from
from hypothesis.strategies import sets

from utilities.hypothesis.tempfile import temp_dirs
from utilities.tempfile import TemporaryDirectory
from utilities.zipfile import yield_zip_file_contents


class TestYieldZipFileContents:
    @given(
        temp_dir=temp_dirs(),
        contents=sets(sampled_from(ascii_letters), min_size=1, max_size=10),
    )
    def test_main(
        self, temp_dir: TemporaryDirectory, contents: set[str]
    ) -> None:
        root = temp_dir.name
        assert root.exists()
        assert not list(root.iterdir())
        path_zip = root.joinpath("zipfile")
        with ZipFile(path_zip, mode="w") as zf:
            for con in contents:
                zf.writestr(con, con)
        assert path_zip.exists()
        with yield_zip_file_contents(path_zip) as paths:
            assert isinstance(paths, list)
            assert len(paths) == len(contents)
            for path in paths:
                assert isinstance(path, Path)



import datetime as dt
from collections.abc import Callable
from enum import Enum as _Enum
from enum import auto

from click import argument
from click import command
from click import echo
from click.testing import CliRunner
from hypothesis import assume
from hypothesis import given
from hypothesis.strategies import DataObject
from hypothesis.strategies import SearchStrategy
from hypothesis.strategies import data
from hypothesis.strategies import dates
from hypothesis.strategies import datetimes
from hypothesis.strategies import just
from hypothesis.strategies import sampled_from
from pytest import mark
from pytest import param

from utilities.click import Date
from utilities.click import DateTime
from utilities.click import Enum
from utilities.click import log_level_option
from utilities.logging import LogLevel


def runners() -> SearchStrategy[CliRunner]:
    return just(CliRunner())


@command()
@argument("date", type=Date())
def uses_date(*, date: dt.date) -> None:
    echo(f"date = {date}")


def format_date_1(date: dt.date, /) -> str:
    return date.isoformat()


def format_date_2(date: dt.date, /) -> str:
    return date.strftime("%4Y%m%d")


class TestDate:
    @given(runner=runners(), date=dates())
    @mark.parametrize("format", [param(format_date_1), param(format_date_2)])
    def test_success(
        self, runner: CliRunner, date: dt.date, format: Callable[[dt.date], str]
    ) -> None:
        result = runner.invoke(uses_date, [format(date)])
        assert result.exit_code == 0
        assert result.stdout == f"date = {date:%4Y-%m-%d}\n"

    @given(runner=runners(), date=dates())
    def test_failure(self, runner: CliRunner, date: dt.date) -> None:
        result = runner.invoke(uses_date, [date.strftime("%4Y/%m/%d")])
        assert result.exit_code == 2


@command()
@argument("datetime", type=DateTime())
def uses_datetime(*, datetime: dt.datetime) -> None:
    echo(f"datetime = {datetime}")


def format_datetime_1(date: dt.datetime, /) -> str:
    return date.isoformat()


def format_datetime_2(date: dt.datetime, /) -> str:
    return date.strftime("%4Y%m%d%H%M%S")


class TestDateTime:
    @given(runner=runners(), date=datetimes())
    @mark.parametrize(
        "format", [param(format_datetime_1), param(format_datetime_2)]
    )
    def test_success(
        self,
        runner: CliRunner,
        date: dt.datetime,
        format: Callable[[dt.datetime], str],
    ) -> None:
        _ = assume(date.microsecond == 0)
        result = runner.invoke(uses_datetime, [format(date)])
        assert result.exit_code == 0
        assert result.stdout == f"datetime = {date}\n"

    @given(runner=runners(), date=dates())
    def test_failure(self, runner: CliRunner, date: dt.date) -> None:
        result = runner.invoke(
            uses_datetime, [date.strftime("%4Y/%m/%d %H:%M:%S")]
        )
        assert result.exit_code == 2


class Truth(_Enum):
    true = auto()
    false = auto()


@command()
@argument("truth", type=Enum(Truth))
def uses_enum(*, truth: Truth) -> None:
    echo(f"truth = {truth}")


class TestEnum:
    @given(data=data(), runner=runners(), truth=sampled_from(Truth))
    def test_success(
        self, data: DataObject, runner: CliRunner, truth: Truth
    ) -> None:
        name = truth.name
        as_str = data.draw(sampled_from([name, name.lower()]))
        result = runner.invoke(uses_enum, [as_str])
        assert result.exit_code == 0
        assert result.stdout == f"truth = {truth}\n"

    @given(runner=runners())
    def test_failure(self, runner: CliRunner) -> None:
        result = runner.invoke(uses_enum, ["not_an_element"])
        assert result.exit_code == 2


@command()
@log_level_option
def uses_log_level(*, log_level: LogLevel) -> None:
    echo(f"log_level = {log_level}")


class TestLogLevelOption:
    @given(runner=runners(), log_level=sampled_from(LogLevel))
    def test_main(self, runner: CliRunner, log_level: LogLevel) -> None:
        result = runner.invoke(uses_log_level, [f"--log-level={log_level}"])
        assert result.exit_code == 0
        assert result.stdout == f"log_level = {log_level}\n"



from pytest import raises

from utilities.text import ensure_str


class TestEnsureStr:
    def test_str(self) -> None:
        assert isinstance(ensure_str(""), str)

    def test_not_str(self) -> None:
        with raises(TypeError, match="x=None"):
            _ = ensure_str(None)



import datetime as dt
from typing import Any
from typing import cast

from hypothesis import assume
from hypothesis import given
from hypothesis.strategies import dates
from hypothesis.strategies import timedeltas
from hypothesis.strategies import times
from pytest import mark
from pytest import raises
from typed_settings import settings
from typed_settings.exceptions import InvalidValueError

from utilities.typed_settings import load_settings


class TestTypedSettings:
    @given(date=dates())
    def test_date(self, date: dt.date) -> None:
        @settings(frozen=True)
        class Config:
            as_date: dt.date = date
            as_str: dt.date = cast(dt.date, date.isoformat())

        config = load_settings(Config, "appname")
        assert isinstance(config, Config)
        assert isinstance(config.as_date, dt.date)
        assert isinstance(config.as_str, dt.date)

    @given(time=times())
    def test_time(self, time: dt.time) -> None:
        @settings(frozen=True)
        class Config:
            as_time: dt.time = time
            as_str: dt.time = cast(dt.time, time.isoformat())

        config = load_settings(Config, "appname")
        assert isinstance(config, Config)
        assert isinstance(config.as_time, dt.time)
        assert isinstance(config.as_str, dt.time)

    @given(
        timedelta=timedeltas(
            min_value=dt.timedelta(0), max_value=dt.timedelta(days=1)
        )
    )
    def test_timedelta(self, timedelta: dt.timedelta) -> None:
        _ = assume(
            (timedelta < dt.timedelta(days=1)) and (timedelta.microseconds == 0)
        )

        @settings(frozen=True)
        class Config:
            as_timedelta: dt.timedelta = timedelta
            as_str: dt.timedelta = cast(dt.timedelta, str(timedelta))

        config = load_settings(Config, "appname")
        assert isinstance(config, Config)
        assert isinstance(config.as_timedelta, dt.timedelta)
        assert isinstance(config.as_str, dt.timedelta)

    @mark.parametrize("cls", [dt.date, dt.time, dt.timedelta])
    def test_errors(self, cls: Any) -> None:
        @settings(frozen=True)
        class Config:
            value: cls = cast(cls, None)

        with raises(InvalidValueError):
            _ = load_settings(Config, "appname")



from string import ascii_letters

from hypothesis import given
from hypothesis.strategies import sampled_from
from hypothesis.strategies import sets

from utilities.hypothesis.tempfile import temp_dirs
from utilities.tempfile import TemporaryDirectory


class TestTempDirs:
    @given(temp_dir=temp_dirs())
    def test_main(self, temp_dir: TemporaryDirectory) -> None:
        path = temp_dir.name
        assert path.is_dir()
        assert len(set(path.iterdir())) == 0

    @given(
        temp_dir=temp_dirs(),
        contents=sets(sampled_from(ascii_letters), max_size=10),
    )
    def test_writing_files(
        self, temp_dir: TemporaryDirectory, contents: set[str]
    ) -> None:
        path = temp_dir.name
        assert len(set(path.iterdir())) == 0
        for content in contents:
            path.joinpath(content).touch()
        assert len(set(path.iterdir())) == len(contents)
